import itertools
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Sequence
from typing import Literal

import tqdm

from musictool import config


def iter_scales(kind, start=None):
    scales = getattr(config, kind)
    it = itertools.cycle(scales)
    if start:
        it = itertools.dropwhile(lambda x: x != start, it)
    it = itertools.islice(it, len(scales))
    return list(it)


# @functools.cache ?
def sequence_builder(
    n: int,
    options: Iterable | Sequence[Iterable] | Callable,
    options_kind: Literal['iterable', 'fixed_per_step', 'callable_from_curr'] = 'iterable',
    curr_prev_constraint: dict[int, Callable] | None = None,
    candidate_constraint: Callable | None = None,
    i_constraints: dict[int, Callable] = dict(),
    unique_key: Callable | None = None,
    loop: bool = False,
    prefix: Sequence = tuple(),
) -> Sequence:
    """build a sequence of elements from options
    Parameters
    ----------
    n : int
        len of returned sequence
    options: Iterable
        on each build step use one of the `options`
    options_kind : str
        'iterable: - todo
        'fixed_per_step: - todo
        'callable_from_curr: - todo
    curr_prev_constraint
    candidate_constraint
    i_constraints
    unique_key: Callable | None
        if not None, sequence elements should have unique keys
    loop : bool
        if True then also checks curr_prev_constraint(seq[-1], seq[0])
    prefix:

    Returns
    -------
    Sequence
    """
    seq = list(prefix) if prefix else list()
    if options_kind == 'iterable':
        options_seq = tuple(options)
        options = frozenset(options_seq)
        if len(options_seq) != len(options):
            raise ValueError('options should be unique')
        ops = options
    elif options_kind == 'fixed_per_step':
        ops = options[len(seq)]
    elif options_kind == 'callable_from_curr':
        assert prefix
        ops = options(seq[-1])
    else:
        raise ValueError('unsupported options_kind')

    if i_constraint := i_constraints.get(len(seq)):
        ops = filter(i_constraint, ops)

    if unique_key:
        prefix_keys = frozenset(unique_key(op) for op in prefix)
        ops = frozenset(op for op in ops if unique_key(op) not in prefix_keys)

    if curr_prev_constraint is not None:

        if abs(max(curr_prev_constraint.keys())) >= n:
            raise IndexError('max index to look back in curr_prev_constraint should be less than n')

        for k, f in curr_prev_constraint.items():
            if abs(k) > len(seq):
                continue
            ops = [op for op in ops if f(seq[k], op)]

    if not prefix:
        ops = tqdm.tqdm(ops)

    for op in ops:
        candidate = seq + [op]
        if candidate_constraint is not None and not candidate_constraint(candidate):
            continue
        if len(candidate) == n:
            if curr_prev_constraint is not None and loop:
                for k, f in curr_prev_constraint.items():
                    if any(not f(candidate[(i + k) % n], candidate[i]) for i in range(abs(k))):
                        break
                else:
                    yield tuple(candidate)
            else:
                yield tuple(candidate)
        else:
            yield from sequence_builder(n, options, options_kind, curr_prev_constraint, candidate_constraint, i_constraints, unique_key, loop, prefix=candidate)



import functools
import itertools
from collections.abc import Callable

from musictool.chord import SpecificChord
from musictool.progression import Progression
from musictool.scale import Scale


def chord_check_cache(f: Callable):
    cache = {}
    cache_info = {'hits': 0, 'misses': 0, 'currsize': 0}

    @functools.wraps(f)
    def inner(a: SpecificChord, b: SpecificChord, *args) -> bool:
        a, b = Progression((a, b)).transposed_to_C0
        key = a, b, *args
        cached = cache.get(key)
        if cached is not None:
            cache_info['hits'] += 1
            return cached
        cache_info['misses'] += 1
        cache_info['currsize'] += 1
        computed = f(a, b, *args)
        cache[key] = computed
        return computed
    inner._cache = cache
    inner._cache_info = cache_info
    return inner


@chord_check_cache
def parallel_interval(a: SpecificChord, b: SpecificChord, interval: int, /) -> bool:
    '''
    parallel in same voices!
    if there'are eg fifth in 1st and fifth in 2nd chord but not from same voices
    - then it allowed (aint considered parallel) (test it)

    a1 - b1
    a0 - b0
    todo: what about fifths + octave (eg C5 G6 -> F5 C6)
    '''
    voice_transitions = tuple(zip(a, b))
    for (a0, b0), (a1, b1) in itertools.combinations(voice_transitions, 2):
        if abs(a0 - a1) % 12 == interval == abs(b0 - b1) % 12:
            return True
    return False


@chord_check_cache
def hidden_parallel(a: SpecificChord, b: SpecificChord, interval: int, /) -> bool:
    """
    hidden/direct parallel/consecutive interval is when:
        1. outer voices (lower and higher) go in same direction (instead of oblique or contrary motion)
        2. they approach param:interval
    voice leading rules often forbid hidden fifths and octaves (param:interval = 7, 0) (explanation: 12 % 12 == 0 octave equal to unison)
    """
    a_low, a_high = a[0], a[-1]
    b_low, b_high = b[0], b[-1]

    is_same_direction = (a_low < b_low and a_high < b_high) or (a_low > b_low and a_high > b_high)
    return is_same_direction and (b_high - b_low) % 12 == interval


@chord_check_cache
def voice_crossing(a: SpecificChord, b: SpecificChord, /) -> bool:
    n = len(b)
    for i in range(n):
        upper = i < n - 1 and b[i] > a[i + 1]
        lower = i > 0 and b[i] < a[i - 1]
        if upper or lower:
            return True
    return False


@chord_check_cache
def large_leaps(a: SpecificChord, b: SpecificChord, interval: int, /) -> bool:
    return any(abs(an - bn) > interval for an, bn in zip(a, b))


def large_spacing(c: SpecificChord, max_interval=12, /):
    return any(b - a > max_interval for a, b in itertools.pairwise(c))


def small_spacing(c: SpecificChord, min_interval=3, /):
    return any(b - a < min_interval for a, b in itertools.pairwise(c))


@chord_check_cache
def make_major_scale_leading_tone_resolving_semitone_up(
    a: SpecificChord,
    b: SpecificChord,
    s: Scale,
    /,
) -> bool:
    if s.name != 'major':
        raise ValueError('not ero')
    leading_tone = [note for note in a.notes if note.abstract == s.notes_ascending[-1]][0]
    tonic = [note for note in b.notes if note.abstract == s.root][0]
    return tonic - leading_tone == 1



from __future__ import annotations

import functools
import itertools

from musictool.chord import SpecificChord
from musictool.note import SpecificNote
from musictool.util.cache import Cached


class Progression(Cached):
    def __init__(self, chords: tuple[SpecificChord, ...], /):
        self.chords = chords
        if not all(isinstance(x, SpecificChord) for x in chords):
            raise TypeError('only SpecificChord items allowed')

    def __getitem__(self, item):
        return self.chords[item]

    def __len__(self):
        return len(self.chords)

    def __repr__(self):
        return f'Progression{self.chords})'

    def all(self, checks__):
        return all(check(a, b) for a, b in itertools.pairwise(self) for check in checks__)

    def all_not(self, checks__):
        return all(not check(a, b) for a, b in itertools.pairwise(self) for check in checks__)

    @property
    def distance(self):
        n = len(self)
        return sum(abs(self[i] - self[(i + 1) % n]) for i in range(n))

    def transpose_unique_key(self, origin_name: bool = True):
        origin = self[0][0]
        key = tuple(frozenset(note - origin for note in chord.notes) for chord in self)
        if origin_name:
            return origin.abstract.i, key
        return key

    def __add__(self, other: int) -> Progression:
        if not isinstance(other, int):
            raise TypeError('only adding integers is allowed (transposition)')
        return Progression(tuple(chord + other for chord in self))

    @functools.cached_property
    def transposed_to_C0(self) -> Progression:
        return self + (SpecificNote('C', 0) - self[0][0])



import random
import string

import cv2
import numpy as np

from musictool import config


def overlay_rect(
    im: np.ndarray,
    pt1: tuple[int, int],
    pt2: tuple[int, int],
    color: tuple[int, int, int],
    alpha: float,
) -> np.ndarray:
    x1, y1 = pt1
    x2, y2 = pt2
    if y2 - y1 > 0 and x2 - x1 > 0:
        sub_im = im[y1:y2, x1:x2]
        rect = np.full_like(sub_im, fill_value=255)
        rect[:, :, 0] = color[0]
        rect[:, :, 1] = color[1]
        rect[:, :, 2] = color[2]
        # print('*********>', x1,x2,y1,y2,alpha, sub_im.shape, rect.shape)
        im[y1:y2, x1:x2] = cv2.addWeighted(rect, alpha, sub_im, 1 - alpha, gamma=0)
    return im


def overlay_image(im: np.ndarray, overlay: np.ndarray, alpha: float,) -> np.ndarray:
    return cv2.addWeighted(overlay, alpha, im, 1 - alpha, gamma=0)


def minmax_scaler(value, oldmin, oldmax, newmin=0.0, newmax=1.0):
    '''
    >>> minmax_scaler(50, 0, 100, 0.0, 1.0)
    0.5
    '''
    return (value - oldmin) * (newmax - newmin) / (oldmax - oldmin) + newmin


def rel_to_abs_w(value): return int(minmax_scaler(value, 0, 1, 0, config.frame_width))
def rel_to_abs_h(value): return int(minmax_scaler(value, 0, 1, 0, config.frame_height))


def rel_to_abs(x, y):
    """xy: coordinates in fractions of screen"""
    return rel_to_abs_w(x), rel_to_abs_h(y)


def random_xy():
    return random.randrange(config.frame_width), random.randrange(config.frame_height)


def random_text(words=tuple(''.join(random.choices(string.ascii_letters, k=random.randint(1, 15))) for _ in range(200))):
    return random.choice(words)



from musictool import config


def bars_to_seconds(
    bars: int | float,
    bpm: int | float = config.beats_per_minute,
    beats_per_bar: int = config.beats_per_bar,
) -> float:
    return bars * 60 * beats_per_bar / bpm



class Cached:
    _cache = {}

    def __new__(cls, *args, **kwargs):
        key = args, frozenset(kwargs.items())
        instance = cls._cache.get(key)
        if instance is not None:
            return instance
        instance = super().__new__(cls)
        instance.__init__(*args, **kwargs)
        cls._cache[key] = instance
        return instance



from __future__ import annotations

import asyncio
import functools
from collections.abc import Iterable
from numbers import Number

from musictool import config
from musictool.midi import player
from musictool.util.cache import Cached


class Note(Cached):
    """
    abstract note, no octave/key
    kinda music theoretic pitch-class
    """

    def __init__(self, name: str):
        """:param name: one of CdDeEFfGaAbB"""
        self.name = name
        self.i = config.note_i[name]
        self.is_black = config.is_black[name]

    @classmethod
    def from_i(cls, i: int):
        return cls(config.chromatic_notes[i % 12])

    def short_repr(self): return self.name
    def __repr__(self): return f'Note(name={self.name})'

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, Note):
            return self.name == other.name

    def __hash__(self): return hash(self.name)

    def __add__(self, other: int):
        return Note.from_i(self.i + other)

    def __sub__(self, other):
        """
        kinda constraint (may be it will be changed later):
            if you computing distance between abstract notes - then self considered above other
            G - C == 7 # C0 G0
            C - G == 5 # G0 C1
        """
        if other.i <= self.i:
            return self.i - other.i
        return 12 + self.i - other.i


@functools.total_ordering
class SpecificNote(Note):
    def __init__(self, abstract: Note | str, octave: int):
        """
        :param octave: in midi format (C5-midi == C3-ableton)
        """
        if isinstance(abstract, str):
            abstract = Note(abstract)
        self.abstract = abstract
        super().__init__(abstract.name)
        self.octave = octave
        self.absolute_i = octave * 12 + self.i  # this is also midi_code
        self.key = self.abstract, self.octave

    @classmethod
    def from_absolute_i(cls, absolute_i: int):
        div, mod = divmod(absolute_i, 12)
        return cls(Note(config.chromatic_notes[mod]), octave=div)

    @classmethod
    def from_str(cls, string: str):
        return cls(Note(string[0]), int(string[1:]))

    async def play(self, seconds: Number = 1):
        player.send_message('note_on', note=self.absolute_i, channel=0)
        await asyncio.sleep(seconds)
        player.send_message('note_off', note=self.absolute_i, channel=0)

    def __repr__(self): return f'{self.abstract.name}{self.octave}'

    def __eq__(self, other):
        if isinstance(other, SpecificNote):
            return self.key == other.key
        elif isinstance(other, str):
            return self.key == SpecificNote.from_str(other).key

    def __hash__(self): return hash(self.key)
    def __lt__(self, other): return self.absolute_i < other.absolute_i

    @functools.cache
    def __sub__(self, other: SpecificNote) -> int:
        """distance between notes"""
        return self.absolute_i - other.absolute_i

    def __add__(self, other: int):
        """C + 7 = G"""
        return SpecificNote.from_absolute_i(self.absolute_i + other)

    @staticmethod
    def to_astract(notes: Iterable[SpecificNote]) -> frozenset[Note]:
        return frozenset(note.abstract for note in notes)


AnyNote = str | Note | SpecificNote


def str_to_note(note: str) -> Note | SpecificNote:
    if len(note) == 0:
        raise ValueError('invalid note string representation')
    elif len(note) == 1:
        return Note(note)
    elif len(note) > 1:
        return SpecificNote.from_str(note)



from setuptools import find_namespace_packages
from setuptools import setup


setup(
    name='musictool',
    version='1.1.25',
    description='set of tools to help learning scales, modes, modulations, chord progressions, voice leading, rhythm',
    long_description_content_type="text/markdown",
    url='https://github.com/tandav/musictool',
    packages=find_namespace_packages(exclude=('static*', 'tests*')),
    install_requires=[
        'mido',
        'pipe21',
    ],
)



from __future__ import annotations

import functools
import itertools
from collections import defaultdict

from musictool import config
from musictool.chord import Chord
from musictool.note import Note
from musictool.noteset import NoteSet
from musictool.piano import Piano
from musictool.util.color import hex_to_rgb
from musictool.util.iteration import iter_scales
from musictool.util.text import cprint


class Scale(NoteSet):
    intervals_to_name = {
        # diatonic
        frozenset({2, 4, 5, 7, 9, 11}): 'major',
        frozenset({2, 3, 5, 7, 9, 10}): 'dorian',
        frozenset({1, 3, 5, 7, 8, 10}): 'phrygian',
        frozenset({2, 4, 6, 7, 9, 11}): 'lydian',
        frozenset({2, 4, 5, 7, 9, 10}): 'mixolydian',
        frozenset({2, 3, 5, 7, 8, 10}): 'minor',
        frozenset({1, 3, 5, 6, 8, 10}): 'locrian',
        # pentatonic
        frozenset({2, 4, 7, 9}): 'p_major',
        frozenset({2, 5, 7, 10}): 'p_dorian',
        frozenset({3, 5, 8, 10}): 'p_phrygian',
        frozenset({2, 5, 7, 9}): 'p_mixolydian',
        frozenset({3, 5, 7, 10}): 'p_minor',
        # sudu
        frozenset({2, 4, 5, 7, 9}): 's_major',
        frozenset({2, 3, 5, 7, 10}): 's_dorian',
        frozenset({1, 3, 5, 8, 10}): 's_phrygian',
        frozenset({2, 4, 7, 9, 11}): 's_lydian',
        frozenset({2, 5, 7, 9, 10}): 's_mixolydian',
        frozenset({3, 5, 7, 8, 10}): 's_minor',
    }
    name_to_intervals = {v: k for k, v in intervals_to_name.items()}

    def __init__(
        self,
        notes: frozenset[str | Note],
        *,
        root: str | Note,
    ):
        super().__init__(notes, root=root)
        self.kind = config.kinds.get(self.name)

        if self.kind is not None:
            self.note_colors = {}
            self.note_scales = {}

            for note, scale in zip(self.notes_ascending, iter_scales(self.kind, start=self.name)):
                self.note_colors[note] = hex_to_rgb(config.scale_colors[scale])
                self.note_scales[note] = scale

            if self.kind == 'diatonic':
                self.triads = self._make_nths(frozenset({0, 2, 4}))
                self.sevenths = self._make_nths(frozenset({0, 2, 4, 6}))
                self.ninths = self._make_nths(frozenset({0, 2, 4, 6, 8}))
                self.notes_to_triad_root = {triad.notes: triad.root for triad in self.triads}
                self.notes_to_seventh_root = {seventh.notes: seventh.root for seventh in self.sevenths}
                self.notes_to_ninth_root = {ninth.notes: ninth.root for ninth in self.ninths}

        self.html_classes = ('card', self.name)

    def _make_nths(self, ns: frozenset[int]) -> tuple[Chord]:
        return tuple(
            Chord(frozenset(self.notes_ascending[(i + n) % len(self)] for n in ns), root=self.notes_ascending[i])
            for i in range(len(self))
        )

    def parallel(self, parallel_name: str | None = None) -> Scale:
        """same root, convert major to minor and vice versa"""
        if parallel_name is None:
            parallel_name = {'major': 'minor', 'minor': 'major'}[self.name]
        return Scale.from_name(self.root, parallel_name)

    def relative(self, relative_name: str | None = None) -> Scale:
        """same set of notes, changes root, convert major to minor and vice versa"""
        if relative_name is None:
            relative_name = {'major': 'minor', 'minor': 'major'}[self.name]
        for note, name in self.note_scales.items():
            if name == relative_name:
                return Scale.from_name(note, name)

    def to_piano_image(self):
        return Piano(scale=self)._repr_svg_()

    def with_html_classes(self, classes: tuple):
        prev = self.html_classes
        self.html_classes = prev + classes
        r = self._repr_html_()
        self.html_classes = prev
        return r

    def _repr_html_(self):
        # <code>bits: {self.bits}</code><br>
        # chords_hover = f"title='{self._chords_text()}'" if self.kind =='diatonic' else ''
        chords_hover = ''
        if C_name := self.note_scales.get(Note('C'), ''):
            C_name = f' | C {C_name}'
        return f'''
        <div class='{' '.join(self.html_classes)}' {chords_hover}>
        <a href='{self.root.name}'><span class='card_header'><h3>{self.root.name} {self.name}{C_name}</h3></span></a>
        {self.to_piano_image()}
        </div>
        '''

# flake8: noqa


class ComparedScales:
    '''
    this is compared scale
    local terminology: left sclae is compared to right
    left is kinda parent, right is kinda child
    '''

    def __init__(self, left: Scale, right: Scale):
        self.left = left
        self.right = right
        self.key = left, right
        self.shared_notes = frozenset(left.notes) & frozenset(right.notes)
        self.new_notes = frozenset(right.notes) - frozenset(left.notes)
        self.del_notes = frozenset(left.notes) - frozenset(right.notes)
        if right.kind == 'diatonic':
            self.shared_triads = frozenset(left.triads) & frozenset(right.triads)
        self.html_classes = ('card',)

    def with_html_classes(self, classes: tuple):
        prev = self.html_classes
        self.html_classes = prev + classes
        r = self._repr_html_()
        self.html_classes = prev
        return r

    # def __format__(self, format_spec): raise No

    # @functools.cached_property
    def _repr_html_(self):
        # <code>bits: {self.bits}</code><br>
        # chords_hover = f"title='{self._chords_text()}'" if self.kind =='diatonic' else ''
        chords_hover = ''
        return f'''
        <div class='{' '.join(self.html_classes)}' {chords_hover}>
        <a href='{self.right.root.name}'><span class='card_header'><h3>{self.right.root.name} {self.right.name}</h3></span></a>
        {self.to_piano_image()}
        </div>
        '''

    def to_piano_image(self, as_base64=False):

        return Piano(
            scale=self.right,
            red_notes=self.del_notes, green_notes=self.new_notes, blue_notes=self.shared_notes,
            notes_squares={
                chord.root: (
                    hex_to_rgb(config.chord_colors[chord.name]),
                    config.BLUE_COLOR if chord in self.shared_triads else config.BLACK_COLOR,
                    config.BLUE_COLOR if chord in self.shared_triads else config.BLACK_COLOR,
                    str(chord),
                )
                for chord in self.right.triads
            } if self.right.kind == 'diatonic' else dict(),
        )._repr_svg_()

    def __eq__(self, other): return self.key == other.key
    def __hash__(self): return hash(self.key)
    def __repr__(self): return f'ComparedScale({self.left.root} {self.left.name} | {self.right.root} {self.right.name})'


diatonic = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.diatonic)}
pentatonic = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.pentatonic)}
sudu = {(root, name): Scale.from_name(root, name) for root, name in itertools.product(config.chromatic_notes, config.sudu)}
all_scales = {'diatonic': diatonic, 'pentatonic': pentatonic, 'sudu': sudu}

# circle of fifths clockwise
majors = dict(
    diatonic=tuple(diatonic[note, 'major'] for note in 'CGDAEBfdaebF'),
    pentatonic=tuple(pentatonic[note, 'p_major'] for note in 'CGDAEBfdaebF'),
    sudu=tuple(sudu[note, 's_major'] for note in 'CGDAEBfdaebF'),
)


@functools.cache
def neighbors(left: Scale):
    neighs = defaultdict(list)
    for right in all_scales[left.kind].values():
        # if left == right:
        #     continue
        right = ComparedScales(left, right)
        neighs[len(right.shared_notes)].append(right)
    return neighs


def print_neighbors(s: Scale):
    neighs = neighbors(s)
    for n_intersect in sorted(neighs.keys(), reverse=True):
        for n in neighs[n_intersect]:
            if n.name != 'major': continue
            print(repr(n).ljust(32), '|', end=' ')
            for note in n.chromatic_mask:
                if note in s.chromatic_mask: print(cprint(note, color='BLUE'), end='')
                else: print(note, end='')
            print(' |', end=' ')
            for chord in n.triads:
                if chord in n.shared_triads: print(cprint(chord, color='BLUE'), end=' ')
                else: print(chord, end=' ')
            print()
        print('=' * 100)

# warm up cache
# for scale in tqdm.tqdm(tuple(itertools.chain(all_scales['diatonic'].values(), all_scales['pentatonic'].values()))):
#     _ = scale.to_piano_image(as_base64=True)
#     for neighbor in itertools.chain.from_iterable(neighbors(scale).values()):
#         _ = neighbor.to_piano_image(as_base64=True)
